import numpy as np

from .control import model_setup
from .cp_confocal import twod, threed
from .cp_triplet import trip
from .cp_mix import double_pnum


# 3D + 2D + T
def CF_Gxyz_3d2dT_gauss(parms, tau):
    u""" Two-component, two- and three-dimensional diffusion
        with a Gaussian laser profile, including a triplet component.
        The triplet factor takes into account blinking according to triplet
        states of excited molecules.
        Set *T* or *τ_trip* to 0, if no triplet component is wanted.

        particle2D = (1-F)/ (1+τ/τ_2D) 
        particle3D = α²*F/( (1+τ/τ_3D) * sqrt(1+τ/(τ_3D*SP²)))
        triplet = 1 + T/(1-T)*exp(-τ/τ_trip)
        norm = (1-F + α*F)²
        G = 1/n*(particle1 + particle2)*triplet/norm + offset

        *parms* - a list of parameters.
        Parameters (parms[i]):
        [0] n       Effective number of particles in confocal volume
                    (n = n2D+n3D)
        [1] τ_2D    Diffusion time of surface bound particls
        [2] τ_3D    Diffusion time of freely diffusing particles
        [3] F       Fraction of molecules of the freely diffusing species
                    (n3D = n*F), 0 <= F <= 1
        [4] SP      SP=z₀/r₀ Structural parameter,
                         describes elongation of the confocal volume
        [5] α       Relative molecular brightness of particle
                    3D compared to particle 2D (α = q3D/q2D)
        [6] τ_trip  Characteristic residence time in triplet state
        [7] T       Fraction of particles in triplet (non-fluorescent) state
                    0 <= T < 1
        [8] offset
        *tau* - lag time
    """
    n = parms[0]
    taud2D = parms[1]
    taud3D = parms[2]
    F = parms[3]
    SP = parms[4]
    alpha = parms[5]
    tautrip = parms[6]
    T = parms[7]
    off = parms[8]

    g = double_pnum(n=n,
                    F1=1-F,
                    alpha=alpha,
                    comp1=twod,
                    comp2=threed,
                    kwargs1={"tau": tau,
                             "taudiff": taud2D},
                    kwargs2={"tau": tau,
                             "taudiff": taud3D,
                             "SP": SP},
                    )

    tr = trip(tau=tau, T=T, tautrip=tautrip)

    G = off + g*tr
    return G


def supplements(parms, countrate=None):
    u"""Supplementary parameters:
        Effective number of freely diffusing particles in 3D solution:
        [9]  n3D = n*F
        Effective number particles diffusing on 2D surface:
        [10] n2D = n*(1-F)
    """
    # We can only give you the effective particle number
    n = parms[0]
    F3d = parms[3]
    Info = list()
    # The enumeration of these parameters is very important for
    # plotting the normalized curve. Countrate must come out last!
    Info.append([u"n3D", n*F3d])
    Info.append([u"n2D", n*(1.-F3d)])
    if countrate is not None:
        # CPP
        cpp = countrate/n
        Info.append([u"cpp [kHz]", cpp])
    return Info


parms = [
    25,      # n
    240,     # taud2D
    0.1,     # taud3D
    0.5,     # F3D
    7,       # SP
    1.0,     # alpha
    0.001,   # tautrip
    0.01,    # T
    0.0      # offset
]

# Boundaries
# strictly positive
boundaries = [[0, np.inf]]*len(parms)
# F
boundaries[3] = [0, .9999999999999]
# T
boundaries[7] = [0, .9999999999999]
boundaries[-1] = [-np.inf, np.inf]


model_setup(
    modelid=6032,
    name="Separate 3D and 2D diffusion with triplet (confocal)",
    comp="T+3D+2D",
    mtype="Confocal (Gaussian) and triplet",
    fctn=CF_Gxyz_3d2dT_gauss,
    par_labels=[
        u"n",
        u"τ_2D [ms]",
        u"τ_3D [ms]",
        u"F_3D",
        u"SP",
        u"\u03b1"+" (q_3D/q_2D)",
        u"τ_trip [ms]",
        u"T",
        u"offset"
    ],
    par_values=parms,
    par_vary=[True, True, True, True, False, False, False, False, False],
    par_boundaries=boundaries,
    par_constraints=[[2, "<", 1], [6, "<", 2]],
    par_hr_labels=[
        u"n",
        u"τ_2D [ms]",
        u"τ_3D [ms]",
        u"F_3D",
        u"SP",
        u"\u03b1"+" (q_3D/q_2D)",
        u"τ_trip [µs]",
        u"T",
        u"offset"
    ],
    par_hr_factors=[
        1.,     # "n",
        1.,     # "τ_2D [ms]",
        1.,     # "τ_3D [ms]",
        1.,     # "F_3D",
        1.,     # "SP",
        1.,     # u"\u03b1"+" (q_3D/q_2D)",
        1000.,  # "τ_trip [µs]",
        1.,     # "T",
        1.      # "offset"
    ],
    supplementary_method=supplements
)



import numpy as np
from .fib4 import dividAndConquer


"""FCS Bulk Correlation Software

    Copyright (C) 2015  Dominic Waithe

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""


def tttr2xfcs(y, num, NcascStart, NcascEnd, Nsub):
    """autocorr, autotime = tttr2xfcs(y,num,10,20)
     Translation into python of:
     Fast calculation of fluorescence correlation data with asynchronous time-correlated single-photon counting.
     Michael Wahl, Ingo Gregor, Matthias Patting, Jorg Enderlein
     """

    dt = np.max(y)-np.min(y)
    y = np.round(y[:], 0)
    numshape = num.shape[0]

    autotime = np.zeros(((NcascEnd+1)*(Nsub+1), 1))
    auto = np.zeros(
        ((NcascEnd+1)*(Nsub+1), num.shape[1], num.shape[1])).astype(np.float64)
    shift = float(0)
    delta = float(1)

    for j in range(0, NcascEnd):

        # Finds the unique photon times and their indices. The division of 'y' by '2' each cycle makes this more likely.

        y, k1 = np.unique(y, 1)
        k1shape = k1.shape[0]

        # Sums up the photon times in each bin.
        cs = np.cumsum(num, 0).T

        # Prepares difference array so starts with zero.
        diffArr1 = np.zeros((k1shape+1))
        diffArr2 = np.zeros((k1shape+1))

        # Takes the cumulative sum of the unique photon arrivals
        diffArr1[1:] = cs[0, k1].reshape(-1)
        diffArr2[1:] = cs[1, k1].reshape(-1)

        # del k1
        # del cs
        num = np.zeros((k1shape, 2))

        # Finds the total photons in each bin. and represents as count.
        # This is achieved because we have the indices of each unique time photon and cumulative total at each point.
        num[:, 0] = np.diff(diffArr1)
        num[:, 1] = np.diff(diffArr2)
        # diffArr1 = [];
        # diffArr2 = [];

        for k in range(0, Nsub):
            shift = shift + delta
            lag = np.round(shift/delta, 0)

            # Allows the script to be sped up.
            if j >= NcascStart:

                # Old method
                # i1= np.in1d(y,y+lag,assume_unique=True)
                # i2= np.in1d(y+lag,y,assume_unique=True)

                # New method, cython
                i1, i2 = dividAndConquer(y, y+lag)

                # If the weights (num) are one as in the first Ncasc round, then the correlation is equal to np.sum(i1)
                i1 = np.where(i1.astype(bool))[0]
                i2 = np.where(i2.astype(bool))[0]

                # Now we want to weight each photon corectly.
                # Faster dot product method, faster than converting to matrix.
                if i1.size and i2.size:
                    auto[(k+(j)*Nsub), :, :] = np.dot((num[i1, :]).T,
                                                      num[i2, :])/delta

            autotime[k+(j)*Nsub] = shift

        # Equivalent to matlab round when numbers are %.5
        y = np.ceil(np.array(0.5*y))
        delta = 2*delta

    for j in range(0, auto.shape[0]):
        auto[j, :, :] = auto[j, :, :]*dt/(dt-autotime[j])
    autotime = autotime/1000000

    # Removes the trailing zeros.
    idauto = np.where(autotime != 0)[0]
    autotime = autotime[idauto]
    auto = auto[idauto, :, :]
    return auto, autotime


def delayTime2bin(dTimeArr, chanArr, chanNum, winInt):

    decayTime = np.array(dTimeArr)
    # This is the point and which each channel is identified.
    decayTimeCh = decayTime[chanArr == chanNum]

    # Find the first and last entry
    firstDecayTime = 0  # np.min(decayTimeCh).astype(np.int32)
    tempLastDecayTime = np.max(decayTimeCh).astype(np.int32)

    # We floor this as the last bin is always incomplete and so we discard photons.
    numBins = np.floor((tempLastDecayTime-firstDecayTime)/winInt)
    lastDecayTime = numBins*winInt

    bins = np.linspace(firstDecayTime, lastDecayTime, int(numBins)+1)

    photonsInBin, jnk = np.histogram(decayTimeCh, bins)

    # bins are valued as half their span.
    decayScale = bins[:-1]+(winInt/2)

    # decayScale =  np.arange(0,decayTimeCh.shape[0])

    return list(photonsInBin), list(decayScale)



"""Module tools - datarange: Let the user choose the disaplyed time range"""
import numpy as np
import wx


# Menu entry name
MENUINFO = ["&Data range",
            "Select an interval of lag times to be used for fitting."]


class SelectChannels(wx.Frame):
    def __init__(self, parent):
        # parent is main frame
        self.parent = parent
        # Get the window positioning correctly
        pos = self.parent.GetPosition()
        pos = (pos[0]+100, pos[1]+100)
        wx.Frame.__init__(self, parent=self.parent, title="Data range selection",
                          pos=pos, style=wx.DEFAULT_FRAME_STYLE | wx.FRAME_FLOAT_ON_PARENT)
        # MYID
        # This ID is given by the parent for an instance of this class
        self.MyID = None
        # Start drawing
        panel = wx.Panel(self)
        self.panel = panel
        # Page
        self.Page = self.parent.notebook.GetCurrentPage()
        self.Calc_init(self.Page)
        text1 = wx.StaticText(panel, label=u"The lag times τ are stored as an " +
                                           u"array of length ")
        self.textend = wx.StaticText(panel, label="%d." % self.lentau)
        text2 = wx.StaticText(panel,
                              label=u"You may wish to confine this array. " +
                                    u"This can be done here.")
        # Spincontrols:
        FlexSpinSizer = wx.FlexGridSizer(rows=2, cols=4, vgap=5, hgap=5)
        FlexSpinSizer.Add(wx.StaticText(panel, label="Channels:"))
        self.spinstart = wx.SpinCtrl(panel, -1, initial=self.left,
                                     min=self.start0, max=self.end0-1)
        FlexSpinSizer.Add(self.spinstart)
        FlexSpinSizer.Add(wx.StaticText(panel, label=" - "))
        self.spinend = wx.SpinCtrl(panel, -1, initial=self.right,
                                   min=self.start0+1, max=self.end0)
        FlexSpinSizer.Add(self.spinend)
        FlexSpinSizer.Add(wx.StaticText(panel, label="Times [ms]:"))
        self.TextTimesStart = wx.StaticText(panel, label="None")
        FlexSpinSizer.Add(self.TextTimesStart)
        FlexSpinSizer.Add(wx.StaticText(panel, label=" - "))
        self.TextTimesEnd = wx.StaticText(panel, label="None")
        FlexSpinSizer.Add(self.TextTimesEnd)
        # Buttons
        btnapply = wx.Button(panel, wx.ID_ANY, 'Apply')
        btnapplyall = wx.Button(panel, wx.ID_ANY, 'Apply to all pages')
        self.ButtonApply = btnapply
        self.ButtonApplyAll = btnapplyall
        self.Bind(wx.EVT_BUTTON, self.OnApply, btnapply)
        self.Bind(wx.EVT_BUTTON, self.OnApplyAll, btnapplyall)
        self.Bind(wx.EVT_SPINCTRL, self.OnChangeChannels, self.spinend)
        self.Bind(wx.EVT_SPINCTRL, self.OnChangeChannels, self.spinstart)
        # Checkbox
        self.fixcheck = wx.CheckBox(panel, -1,
                                    label="Fix current channel selection for all pages.")
        self.Bind(wx.EVT_CHECKBOX, self.OnCheckbox, self.fixcheck)
        # Text
        channelsel = "Leave this window open for a fixed selection."
        text3 = wx.StaticText(panel, label=channelsel)
        # Sizer
        topSizer = wx.BoxSizer(wx.VERTICAL)
        buttonsizer = wx.BoxSizer(wx.HORIZONTAL)
        buttonsizer.Add(btnapply, 1)
        buttonsizer.Add(btnapplyall, 1)
        text1sizer = wx.BoxSizer(wx.HORIZONTAL)
        text1sizer.Add(text1)
        text1sizer.Add(self.textend)
        topSizer.Add(text1sizer)
        topSizer.Add(text2)
        topSizer.AddSpacer(5)
        topSizer.Add(FlexSpinSizer)
        topSizer.Add(self.fixcheck)
        topSizer.Add(text3)
        topSizer.AddSpacer(5)
        topSizer.Add(buttonsizer)
        panel.SetSizer(topSizer)
        topSizer.Fit(self)
        self.SetMinSize(topSizer.GetMinSize())
        # Get times.
        self.OnChangeChannels()
        # Icon
        if parent.MainIcon is not None:
            wx.Frame.SetIcon(self, parent.MainIcon)
        # Show window
        self.Show(True)
        self.OnPageChanged(self.Page)

    def Calc_init(self, parent):
        # Variables
        # Parent should be the fitting panel -
        # The tab, where the fitting is done.
        self.Page = parent
        if self.Page == None:
            # dummy info
            taufull = np.arange(100)
            self.left = self.right = None
            self.panel.Disable()
        else:
            self.left = self.Page.corr.fit_ival[0]     # starting position
            self.right = self.Page.corr.fit_ival[1]      # ending position
            taufull = self.Page.corr.lag_time
        self.lentau = len(taufull)
        self.start0 = 0                     # left border of interval
        # The interval starts at 0!
        self.end0 = self.lentau - 1         # right border of interval
        if self.left is None or self.left > self.end0:
            # This means, that either left = right = None
            # or the correlation-array is too small
            self.left = self.start0
        if self.right is None:
            # set the maximum possible value
            self.right = self.end0
        else:
            self.right -= 1

    def OnApply(self, event=None):
        self.SetValues()
        self.Page.PlotAll()

    def OnApplyAll(self, event=None):
        N = self.parent.notebook.GetPageCount()
        for i in np.arange(N):
            # Set Page
            Page = self.parent.notebook.GetPage(i)
            # Find out maximal length
            self.SetValues(page=Page)
            Page.PlotAll()
        # Page.PlorAll() calls this function. This results in the wrong data
        # being displayed in an open "View Info" Window. We call it again.
        self.parent.OnFNBPageChanged()
        if self.parent.MenuAutocloseTools.IsChecked():
            # Autoclose
            self.OnClose()

    def OnChangeTimes(self, e=None):
        """ Called, whenever data range in seconds is changed. This updates
            the data range in channels in the window.
            This function might be used in later versions of PyCorrFit.
        """
        pass

    def OnChangeChannels(self, e=None):
        """ Called, whenever data range in channels is changed. This updates
            the data range in seconds in the window.
        """
        if self.Page == None:
            return
        N = len(self.Page.corr.lag_time)
        start = self.spinstart.Value
        end = self.spinend.Value
        # If the initial boundaries are outside of the experimental
        # data array of length N, change the start and end variables.
        start = start*(start < N-2)
        end = min(end, N-1)
        t1 = 1.*self.Page.corr.lag_time[start]
        t2 = 1.*self.Page.corr.lag_time[end]
        self.TextTimesStart.SetLabel("%.4e" % t1)
        self.TextTimesEnd.SetLabel("%.4e" % t2)
        self.OnCheckbox()

    def OnCheckbox(self, event=None):
        """ Set the correct value in the spincontrol, if the checkbox
            is not checked.
        """
        state = self.fixcheck.GetValue()
        if state == True:
            self.OnApplyAll()
            self.ButtonApply.Disable()
            self.ButtonApplyAll.Disable()
        else:
            self.ButtonApply.Enable()
            self.ButtonApplyAll.Enable()
        # self.OnPageChanged(self.Page)

    def OnClose(self, event=None):
        self.parent.toolmenu.Check(self.MyID, False)
        self.parent.ToolsOpen.__delitem__(self.MyID)
        self.Destroy()

    def OnPageChanged(self, page, trigger=None):
        # We do not need the *Range* Commands here yet.
        # We open and close the SelectChannelsFrame every time we
        # import some data.
        #
        if trigger in ["parm_batch", "fit_batch", "page_add_batch"]:
            return
        # Check if we have a fixed channel selection
        if self.parent.notebook.GetPageCount() == 0:
            self.panel.Disable()
        else:
            self.panel.Enable()
            # There is a page. We may continue.
            state = self.fixcheck.GetValue()
            if state:
                # We do not need to run Calc_init
                self.SetValues(page=page)
                page.PlotAll(event="init")
            else:
                # We will run it
                self.Calc_init(page)
                self.spinstart.SetRange(self.start0, self.end0-1)
                self.spinstart.SetValue(self.left)
                self.spinend.SetRange(self.start0+1, self.end0)
                self.spinend.SetValue(self.right)
                self.textend.SetLabel("%d." % self.lentau)
                self.OnChangeChannels()

    def SetValues(self, page=None):
        if page is None:
            page = self.Page
        # Get interval
        start = self.spinstart.GetValue()
        end = self.spinend.GetValue() + 1  # +1, [sic]
        if start > end:
            # swap the variables, we are not angry at the user
            start, end = end, start
        # Find out maximal length
        maxlen = len(page.corr.lag_time)
        # Use the smaller one of both, so we do not get an
        # index out of bounds error
        page.corr.fit_ival = [start*(start < maxlen - 1),
                              min(end, maxlen)
                              ]



import numpy as np

from .control import model_setup
from .cp_confocal import twod, threed
from .cp_triplet import trip
from .cp_mix import triple_pnum


def CF_Gxyz_gauss_3D3D2DT(parms, tau):
    u""" Two three-dimensional and one two-dimensional free diffusion
        with a Gaussian laser profile, including a triplet component.
        The triplet factor takes into account a blinking term.
        Set *T* or *τ_trip* to 0, if no triplet component is wanted.

        F₃ = 1-F₁-F₂
        particle1 =        F₁/( (1+τ/τ₁) * sqrt(1+τ/(τ₁*SP²)))
        particle2 = α₂₁² * F₂/( (1+τ/τ₂) * sqrt(1+τ/(τ₂*SP²)))
        particle3 = α₃₁² * F₃/( (1+τ/τ₃))
        triplet = 1 + T/(1-T)*exp(-τ/τ_trip)
        norm = (F₁ + α₂₁*F₂ + α₃₁*F₃)²
        G = 1/n*(particle1 + particle2 + particle3)*triplet/norm + offset

        *parms* - a list of parameters.
        Parameters (parms[i]):
        [0]  n      Effective number of particles in confocal volume
                    (n = n₁+n₂+n₃)
        [1]  τ₁     Diffusion time of particle species 1 (3D)
        [2]  τ₂     Diffusion time of particle species 2 (3D)
        [3]  τ₃     Diffusion time of particle species 3 (2D)
        [4]  F₁     Fraction of molecules of species 1 (n₁ = n*F₁)
                    0 <= F₁ <= 1
        [5]  F₂     Fraction of molecules of species 2 (n₂ = n*F₂)
                    0 <= F₂ <= 1
        [6]  SP     SP=z₀/r₀, Structural parameter,
                    describes elongation of the confocal volume
        [7]  α₂₁    Relative molecular brightness of particle
                    2 compared to particle 1 (α = q₂/q₁)
        [8]  α₃₁    Relative molecular brightness of particle
                    3 compared to particle 1 (α = q₃/q₁)
        [9]  τ_trip Characteristic residence time in triplet state
        [10] T      Fraction of particles in triplet (non-fluorescent) state
                    0 <= T < 1
        [11] offset
        *tau* - lag time
    """
    n = parms[0]
    taud1 = parms[1]
    taud2 = parms[2]
    taud3 = parms[3]
    F1 = parms[4]
    F2 = parms[5]
    SP = parms[6]
    alpha21 = parms[7]
    alpha31 = parms[8]
    tautrip = parms[9]
    T = parms[10]
    off = parms[11]

    g = triple_pnum(n=n,
                    F1=F1,
                    F2=F2,
                    alpha21=alpha21,
                    alpha31=alpha31,
                    comp1=threed,
                    comp2=threed,
                    comp3=twod,
                    kwargs1={"tau": tau,
                             "taudiff": taud1,
                             "SP": SP},
                    kwargs2={"tau": tau,
                             "taudiff": taud2,
                             "SP": SP},
                    kwargs3={"tau": tau,
                             "taudiff": taud3},
                    )

    tr = trip(tau=tau, T=T, tautrip=tautrip)

    G = off + g*tr
    return G


def supplements(parms, countrate=None):
    u"""Supplementary parameters:
        [12]  n₁ = n*F₁     Particle number of species 1 (3D)
        [13]  n₂ = n*F₂     Particle number of species 2 (3D)
        [14]  n₃ = n*F₃     Particle number of species 3 (2D; F₃ = 1-F₁-F₂)
    """
    # We can only give you the effective particle number
    n = parms[0]
    F1 = parms[4]
    F2 = parms[5]
    Info = list()
    # The enumeration of these parameters is very important for
    # plotting the normalized curve. Countrate must come out last!
    Info.append([u"n\u2081", n*F1])
    Info.append([u"n\u2082", n*F2])
    Info.append([u"n\u2083", n*(1-F1-F2)])
    if countrate is not None:
        # CPP
        cpp = countrate/n
        Info.append(["cpp [kHz]", cpp])
    return Info


parms = [25,      # n
         5,       # taud1
         1000,    # taud2
         11000,   # taud3
         0.5,     # F1
         0.01,    # F2
         5,       # SP
         1.0,     # alpha21
         1.0,     # alpha31
         0.001,   # tautrip
         0.01,    # T
         0.0      # offset
         ]

# Boundaries
# strictly positive
boundaries = [[0, np.inf]]*len(parms)
# F
boundaries[4] = [0, .9999999999999]
boundaries[5] = [0, .9999999999999]
# T
boundaries[10] = [0, .9999999999999]
boundaries[-1] = [-np.inf, np.inf]


model_setup(
    modelid=6082,
    name="Twofold 3D and one 2D diffusion with triplet (confocal)",
    comp="T+3D+3D+2D",
    mtype="Confocal (Gaussian) and triplet",
    fctn=CF_Gxyz_gauss_3D3D2DT,
    par_labels=[
        u"n",
        u"τ"+u"\u2081"+" [ms]",
        u"τ"+u"\u2082"+" [ms]",
        u"τ"+u"\u2083"+" [ms]",
        u"F"+u"\u2081",
        u"F"+u"\u2082",
        u"SP",
        u"\u03b1\u2082\u2081",
        u"\u03b1\u2083\u2081",
        u"τ_trip [ms]",
        u"T",
        u"offset"
    ],
    par_values=parms,
    par_vary=[True, True, False, True,
              False, True, False, False,
              False, False, True, False],
    par_boundaries=boundaries,
    par_constraints=[[2, ">", 1], [3, ">", 2], [9, "<", 1], [5, 4, "<", "1"]],
    par_hr_labels=[
        u"n",
        u"τ"+u"\u2081"+" [ms]",
        u"τ"+u"\u2082"+" [ms]",
        u"τ"+u"\u2083"+" [ms]",
        u"F"+u"\u2081",
        u"F"+u"\u2082",
        u"SP",
        u"\u03b1\u2082\u2081",
        u"\u03b1\u2083\u2081",
        u"τ_trip [µs]",
        u"T",
        u"offset"
    ],
    par_hr_factors=[
        1.,     # n
        1.,     # taud1
        1.,     # taud2
        1.,     # taud3
        1.,     # F1
        1.,     # F2
        1.,     # SP
        1.,     # alpha21
        1.,     # alpha31
        1000.,  # tautrip [µs]
        1.,     # T
        1.      # offset
    ],
    supplementary_method=supplements
)



"""Test support for FCS file formats"""

import os
from os.path import split
import warnings

import numpy as np
import pytest

import data_file_dl
import pycorrfit

# Files that are known to not work
exclude = []

NOAPITOKEN = "GITHUB_API_TOKEN" not in os.environ


@pytest.mark.xfail(NOAPITOKEN, reason="Restrictions to GitHub API")
def test_asc_all_open():
    # get list of supported file extensions
    ext = "ASC"
    files = data_file_dl.get_data_files_ext(ext)
    assert files
    for f in files:
        if [ex for ex in exclude if f.endswith(ex)]:
            continue
        dn, fn = split(f)
        data = pycorrfit.readfiles.open_any(dn, fn)
        assert data


@pytest.mark.xfail(NOAPITOKEN, reason="Restrictions to GitHub API")
def test_asc_alv7004usb():
    """Test alv7004/USB format"""
    f1 = data_file_dl.get_data_file("ALV-7004USB_ac01_cc01_10.ASC")
    data = pycorrfit.readfiles.open_any(f1)
    assert data["Type"] == ["AC1", "AC2", "CC12", "CC21"]
    assert np.allclose(data["Correlation"][0][10], np.array([0.000275, 0.11208]))
    assert np.allclose(data["Correlation"][1][12], np.array([0.000325, 0.0900233]))
    assert np.allclose(data["Correlation"][2][18], np.array([0.00055, 0.0582773]))
    assert np.allclose(data["Correlation"][3][120], np.array([3.6864, 0.0224212]))
    assert len(data["Trace"][0]) == 253
    assert len(data["Trace"][1]) == 253
    assert len(data["Trace"][2]) == 2
    assert len(data["Trace"][3]) == 2
    assert np.all(data["Trace"][0] == data["Trace"][2][0])
    assert np.all(data["Trace"][1] == data["Trace"][2][1])
    assert np.all(data["Trace"][0] == data["Trace"][3][0])
    assert np.all(data["Trace"][1] == data["Trace"][3][1])
    assert np.allclose(data["Trace"][0][10], np.array([1289.06, 140.20404]))
    assert np.allclose(data["Trace"][1][100], np.array([11835.94, 94.68225]))

    f2 = data_file_dl.get_data_file("ALV-7004USB_dia10_cen10_0001.ASC")
    data2 = pycorrfit.readfiles.open_any(f2)
    # There are empty AC2 and CC12/CC21 curves in this file that should
    # be removed by pycorrfit.
    assert data2["Type"] == ["AC1"]
    assert np.allclose(data2["Correlation"][0][56], np.array([0.0144, 0.0513857]))
    assert len(data2["Trace"][0]) == 254
    assert np.allclose(data2["Trace"][0][210], np.array([49453.13, 165.41434]))

    f3 = data_file_dl.get_data_file("ALV-7004.ASC")
    data3 = pycorrfit.readfiles.open_any(f3)
    assert len(data3["Type"]) == 1
    assert len(data3["Trace"][0]) == 66
    assert data3["Type"][0] == "AC"
    assert np.allclose(data3["Correlation"][0][56], np.array([0.0144, 0.38757]))
    assert np.allclose(data3["Trace"][0][60], np.array([1.21523440e5, 5.11968700e1]))

    f4 = data_file_dl.get_data_file("ALV-7004USB_ac3.ASC")
    data4 = pycorrfit.readfiles.open_any(f4)
    assert len(data4["Type"]) == 1
    assert data4["Type"][0] == "AC"
    assert len(data4["Trace"][0]) == 254


@pytest.mark.xfail(NOAPITOKEN, reason="Restrictions to GitHub API")
def test_cor_all_open():
    f1 = data_file_dl.get_data_file("1_nM_Great_Correlation_Curve_121622.cor")
    data = pycorrfit.readfiles.open_any(f1)
    assert len(data["Filename"]) == 3
    assert data["Filename"][0] == "1_nM_Great_Correlation_Curve_121622.cor"
    assert len(data["Trace"]) == 3
    assert data["Trace"][0] == []
    assert data["Type"] == ["AC", "CC", "AC"]
    assert len(data["Correlation"]) == 3
    assert all(map(lambda x: x.shape == (145, 2), data["Correlation"]))


@pytest.mark.xfail(NOAPITOKEN, reason="Restrictions to GitHub API")
def test_csv_all_open():
    # get list of supported file extensions
    ext = "csv"
    files = data_file_dl.get_data_files_ext(ext)
    assert files
    for f in files:
        if [ex for ex in exclude if f.endswith(ex)]:
            continue
        dn, fn = split(f)
        data = pycorrfit.readfiles.open_any(dn, fn)
        assert data


@pytest.mark.xfail(NOAPITOKEN, reason="Restrictions to GitHub API")
def test_fcs_all_open():
    # get list of supported file extensions
    ext = "fcs"
    files = data_file_dl.get_data_files_ext(ext)
    assert files
    for f in files:
        if [ex for ex in exclude if f.endswith(ex)]:
            continue
        dn, fn = split(f)
        data = pycorrfit.readfiles.open_any(dn, fn)
        assert data


@pytest.mark.xfail(NOAPITOKEN, reason="Restrictions to GitHub API")
def test_fcs_channel_names():
    """Test arbitrary channel names (instead of just "1" and "2"

    The file '001_A488_channel_names.fcs' was created manually.
    """
    fref = data_file_dl.get_data_file("Zeiss_Confocor3_A488+GFP/001_A488.fcs")
    fed = data_file_dl.get_data_file("001_A488_channel_names.fcs")

    dataref = pycorrfit.readfiles.open_any(fref)
    dataed = pycorrfit.readfiles.open_any(fed)
    assert np.all(dataref["Correlation"][0] == dataed["Correlation"][0])
    assert np.all(dataref["Correlation"][1] == dataed["Correlation"][1])
    assert np.all(dataref["Correlation"][2] == dataed["Correlation"][2])
    assert np.all(dataref["Correlation"][3] == dataed["Correlation"][3])
    assert np.all(dataref["Trace"][0] == dataed["Trace"][0])
    assert np.all(dataref["Trace"][1] == dataed["Trace"][1])
    assert np.all(dataref["Trace"][2][0] == dataed["Trace"][2][0])
    assert np.all(dataref["Trace"][2][1] == dataed["Trace"][2][1])
    assert np.all(dataref["Trace"][3][0] == dataed["Trace"][3][0])
    assert np.all(dataref["Trace"][3][1] == dataed["Trace"][3][1])
    assert dataref["Type"] == dataed["Type"]
    assert dataref["Filename"] != dataed["Filename"]


@pytest.mark.xfail(NOAPITOKEN, reason="Restrictions to GitHub API")
def test_pt3_all_open():
    # get list of supported file extensions
    ext = "pt3"
    files = data_file_dl.get_data_files_ext(ext)
    assert files
    for f in files:
        if [ex for ex in exclude if f.endswith(ex)]:
            continue
        dn, fn = split(f)
        data = pycorrfit.readfiles.open_any(dn, fn)
        assert data


@pytest.mark.xfail(NOAPITOKEN, reason="Restrictions to GitHub API")
def test_pt3_basic():
    f1 = data_file_dl.get_data_file("PicoQuant_SymphoTime32_A42F-4jul2014/Point_1.pt3")
    data = pycorrfit.readfiles.open_any(f1)

    trace = data["Trace"][0][0]
    assert trace.shape == (600, 2)
    try:
        assert np.allclose(trace[40], np.array([2037, 6.48]))
    except AssertionError:
        warnings.warn("Unknown pt3 problem after migration to Python3!")

    corr = data["Correlation"][0]
    assert corr.shape == (150, 2)
    assert np.allclose(corr[40], np.array([0.000698, 0.58007174877053136]))
    assert np.allclose(corr[100], np.array([0.72089, 0.019201608388821567]))


@pytest.mark.xfail(NOAPITOKEN, reason="Restrictions to GitHub API")
def test_ptu_all_open():
    ext = "ptu"
    files = data_file_dl.get_data_files_ext(ext)
    assert files
    for f in files:
        if [ex for ex in exclude if f.endswith(ex)]:
            continue
        data = pycorrfit.readfiles.open_any(f)
        assert data


@pytest.mark.xfail(NOAPITOKEN, reason="Restrictions to GitHub API")
def test_ptu_hydraharp_t2_basic():
    f1 = data_file_dl.get_data_file("hydraharp/v20_t2.ptu")
    data = pycorrfit.readfiles.open_any(f1)

    trace = data["Trace"][0][0]
    assert trace.shape == (199, 2)

    corr = data["Correlation"][0]
    assert corr.shape == (150, 2)


@pytest.mark.xfail(NOAPITOKEN, reason="Restrictions to GitHub API")
def test_ptu_hydraharp_t3_basic():
    f1 = data_file_dl.get_data_file("hydraharp/v20_t3.ptu")
    data = pycorrfit.readfiles.open_any(f1)

    trace = data["Trace"][0][0]
    assert trace.shape == (399, 2)

    corr = data["Correlation"][0]
    assert corr.shape == (150, 2)


@pytest.mark.xfail(NOAPITOKEN, reason="Restrictions to GitHub API")
def test_ptu_picoharp_v30_t2_basic():
    f1 = data_file_dl.get_data_file("picoharp/v30_t2.ptu")
    data = pycorrfit.readfiles.open_any(f1)

    trace = data["Trace"][0][0]
    assert trace.shape == (306, 2)

    corr = data["Correlation"][0]
    assert corr.shape == (150, 2)


@pytest.mark.xfail(NOAPITOKEN, reason="Restrictions to GitHub API")
def test_sin_all_open():
    # get list of supported file extensions
    ext = "sin"
    files = data_file_dl.get_data_files_ext(ext)
    for f in files:
        if [ex for ex in exclude if f.endswith(ex)]:
            continue
        dn, fn = split(f)
        data = pycorrfit.readfiles.open_any(dn, fn)
        assert data


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in sorted(list(loc.keys())):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



"""meta data and methods for PyCorrFit"""
import os
import sys


def find_program(program):
    """ Uses the systems PATH variable find executables"""
    path = os.environ['PATH']
    paths = path.split(os.pathsep)
    for d in paths:
        if os.path.isdir(d):
            fullpath = os.path.join(d, program)
            if sys.platform[:3] == 'win':
                for ext in '.exe', '.bat':
                    program_path = fullpath + ext
                    if os.path.isfile(fullpath + ext):
                        return (1, program_path)
            else:
                if os.path.isfile(fullpath):
                    return (1, fullpath)
    return (0, None)


def get_file_location(filename):
    """
    Locate non-Python files that are part of PyCorrFit.
    """
    dirname = os.path.dirname(os.path.abspath(__file__))
    locations = ["/./", "/pycorrfit_doc/", "/doc/"]
    locations += ["/.."+l for l in locations]
    locations = [os.path.realpath(dirname+l) for l in locations]

    for i in range(len(locations)):
        # check /usr/lib64/32 -> /usr/lib
        for larch in ["lib32", "lib64"]:
            if dirname.count(larch):
                locations.append(locations[i].replace(larch, "lib", 1))

    # freezed binaries:
    if hasattr(sys, 'frozen'):
        try:
            adir = sys._MEIPASS + "/doc/"  # @UndefinedVariable
        except:
            adir = "./"
        locations.append(os.path.realpath(adir))
    for loc in locations:
        thechl = os.path.join(loc, filename)
        if os.path.exists(thechl):
            return thechl
            break
    # if this does not work:
    return None



"""
Download experimental data files from https://github.com/FCS-analysis/FCSdata

This module establishes
"""
import json
import os
from os.path import abspath, dirname, join, exists
import pathlib

import urllib3
import warnings

# Download path root
raw_origin = "https://github.com/FCS-analysis/FCSdata/raw/master/"
# GitHub API root
api_origin = "https://api.github.com/repos/FCS-analysis/FCSdata/git/"
# Download directory
dldir = join(dirname(abspath(__file__)), "data")
# Pool Manager handles all requests
pool_manager = urllib3.PoolManager()

_fcs_data_tree = None


def dl_file(url, dest, chunk_size=6553,
            http=pool_manager):
    """
    Download `url` to `dest`.

    Parameters
    ----------
    url : str
        Full download URL
    dest : str
        Full download path. Directory will be created if non-existent.
    chunk_size : int
        Chunk size of download (download buffer size).
    http : instance of `urllib3.PoolManager`
        Manages all connections. Must implement the
        `request` method.
    """
    if not exists(dirname(dest)):
        os.makedirs(dirname(dest))
    r = http.request('GET', url, preload_content=False)
    with open(dest, 'wb') as out:
        while True:
            data = r.read(chunk_size)
            if data is None or len(data) == 0:
                break
            out.write(data)


def get_data_file(filename, dldir=dldir, pool_manager=pool_manager,
                  api_origin=api_origin, raw_origin=raw_origin):
    """
    Return first occurrence of `filename` in the data file tree.
    """
    _f, ext = os.path.splitext(filename)
    assert ext != "", "filename has no extension!"
    extp = ext.strip(".").lower()

    fbase = filename
    fdir = os.path.join(dldir, extp)
    fpath = os.path.join(fdir, fbase)
    if not os.path.exists(fpath):
        # download file if it does not exist
        url = raw_origin+extp+"/"+fbase
        dl_file(url, fpath)

    return fpath


def get_data_files_ext(extension, dldir=dldir, pool_manager=pool_manager,
                       api_origin=api_origin, raw_origin=raw_origin):
    """
    Get all files in the repository `origin` that are
    in the folder `extension` and have a file-ending
    that matches `extension` (case-insensitive).

    The files are downloaded and local paths in the
    `dldir` directory are returned. If no internet
    connection is available, previously downloaded
    files will be used.

    Parameters
    ----------
    extension : str
        A file extension such as `fcs` or `sin`.
    dldir : str
        Path to download directory.
    http : instance of `urllib3.PoolManager`
        Manages all connections. Must implement the
        `request` method.
    raw_origin : str
        Web root for downloads, e.g.
        "https://raw.github.com/FCS-analysis/FCSdata"
    api_origin : str
        GitHub api URL, e.g.
        "https://api.github.com/repos/FCS-analysis/FCSdata/git/"


    Notes
    -----
    The files in the remote location must be sorted according to
    file extionsion. E.g. all `*.sin` files must be located in a
    folder in the root directory named `sin`.
    """
    ext = extension.lower()
    if not ext.startswith("."):
        ext = "."+ext
    try:
        # Get file list and download
        files = get_data_tree_remote(
            pool_manager=pool_manager, api_origin=api_origin)
        # also grab the local files that are in development
        files += list(map(
            lambda p: str(p.relative_to(pathlib.Path(dldir))), 
            filter(
                lambda p: p.is_file(),  
                pathlib.Path(dldir).rglob(f"*{ext}"))))
        extfiles = [f for f in files if f.lower().startswith(
            ext[1:]+"/") and f.lower().endswith(ext)]
        extfiles = list(set(extfiles))
        extfiles.sort()

        dl_files = []

        for f in extfiles:
            dest = join(dldir, f)
            if not exists(dest):
                dl_file(join(raw_origin, f), dest)
            dl_files.append(dest)

    except (urllib3.exceptions.MaxRetryError, KeyError):
        # e.g. no internet connection
        warnings.warn(
            "No connection, using previuously downloaded files only.")
        files = get_data_tree_local(dldir=dldir)
        dl_files = [f for f in files if f.lower().endswith(ext)]

    return dl_files


def get_data_tree_local(dldir=dldir):
    """
    Returns FCSdata repository tree of local files.
    """
    loc_files = []
    for root, _dirs, files in os.walk(dldir):
        # sorting convention: the folder names the extension
        relpath = os.path.relpath(root, dldir)
        ext = os.path.basename(relpath[::-1])[::-1]
        for f in files:
            if f.lower().endswith(ext):
                loc_files.append(os.path.join(root, f))

    return loc_files


def get_data_tree_remote(pool_manager=pool_manager, api_origin=api_origin):
    """
    Returns FCSdata repository tree.
    The tree is saved in the global variable `_fcs_data_tree` to reduce
    number of GitHub API requests.
    """
    global _fcs_data_tree
    if _fcs_data_tree is None:
        url = api_origin+"trees/master?recursive=1"
        # headers
        headers = {'User-Agent': __file__}
        # GitHub API token to prevent rate-limits
        # Key is generated with
        #
        #    gem install travis
        #    travis encrypt GITHUB_API_TOKEN=secret-token
        #
        # Add the result to env in travis.yml.
        if "GITHUB_API_TOKEN" in os.environ:
            headers["Authorization"] = "token {}".format(
                os.environ["GITHUB_API_TOKEN"])
            r = pool_manager.request("GET", url, headers=headers, retries=10)
            jd = json.loads(r.data)
            tree = jd["tree"]
        else:
            r = pool_manager.request("GET", url, headers=headers, retries=10)
            jd = json.loads(r.data)
            tree = jd["tree"]
        fcs_data_tree = [t["path"] for t in tree]
    return fcs_data_tree
